############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Classes representing variables containing quantum types (i.e. Qubits)."""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Iterator, Optional, Sequence, Union

from openpulse import ast
from openpulse.printer import dumps

from oqpy.base import AstConvertible, Var, make_annotations, to_ast
from oqpy.classical_types import _ClassicalVar

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["Qubit", "QubitArray", "defcal", "PhysicalQubits", "Cal"]


class Qubit(Var):
    """OQpy variable representing a single qubit."""

    def __init__(
        self,
        name: str,
        needs_declaration: bool = True,
        annotations: Sequence[str | tuple[str, str]] = (),
    ):
        super().__init__(name, needs_declaration=needs_declaration)
        self.name = name
        self.annotations = annotations

    def to_ast(self, prog: Program) -> ast.Expression:
        """Converts the OQpy variable into an ast node."""
        prog._add_var(self)
        return ast.Identifier(self.name)

    def make_declaration_statement(self, program: Program) -> ast.Statement:
        """Make an ast statement that declares the OQpy variable."""
        decl = ast.QubitDeclaration(ast.Identifier(self.name), size=None)
        decl.annotations = make_annotations(self.annotations)
        return decl


class PhysicalQubits:
    """Provides a means of accessing qubit variables corresponding to physical qubits.

    For example, the openqasm qubit "$3" is accessed by ``PhysicalQubits[3]``.
    """

    def __class_getitem__(cls, item: int) -> Qubit:
        return Qubit(f"${item}", needs_declaration=False)


# Todo (#51): support QubitArray
class QubitArray:
    """Represents an array of qubits."""


@contextlib.contextmanager
def defcal(
    program: Program,
    qubits: Union[Qubit, list[Qubit]],
    name: str,
    arguments: Optional[list[AstConvertible]] = None,
    return_type: Optional[ast.ClassicalType] = None,
) -> Union[Iterator[None], Iterator[list[_ClassicalVar]], Iterator[_ClassicalVar]]:
    """Context manager for creating a defcal.

    .. code-block:: python

        with defcal(program, q1, "X", [AngleVar(name="theta"), oqpy.pi/2], oqpy.bit) as theta:
            program.play(frame, waveform)
    """
    if isinstance(qubits, Qubit):
        qubits = [qubits]
    assert return_type is None or isinstance(return_type, ast.ClassicalType)

    arguments_ast = []
    variables = []
    if arguments is not None:
        for arg in arguments:
            if isinstance(arg, _ClassicalVar):
                arguments_ast.append(
                    ast.ClassicalArgument(type=arg.type, name=ast.Identifier(name=arg.name))
                )
                arg._needs_declaration = False
                variables.append(arg)
            else:
                arguments_ast.append(to_ast(program, arg))

    program._push()
    if len(variables) > 1:
        yield variables
    elif len(variables) == 1:
        yield variables[0]
    else:
        yield None
    state = program._pop()

    stmt = ast.CalibrationDefinition(
        ast.Identifier(name),
        arguments_ast,
        [ast.Identifier(q.name) for q in qubits],
        return_type,
        state.body,
    )
    program._add_statement(stmt)
    program._add_defcal(
        [qubit.name for qubit in qubits], name, [dumps(a) for a in arguments_ast], stmt
    )


@contextlib.contextmanager
def Cal(program: Program) -> Iterator[None]:
    """Context manager that begins a cal block."""
    program._push()
    yield
    state = program._pop()
    program._add_statement(ast.CalibrationStatement(state.body))



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Base classes and conversion methods for OQpy.

This class establishes how expressions are represented in oqpy and how
they are converted to AST nodes.
"""

from __future__ import annotations

import math
import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Iterable, Sequence, Union

import numpy as np
from openpulse import ast

from oqpy import classical_types

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from oqpy import Program


class OQPyExpression:
    """Base class for OQPy expressions.

    Subclasses must implement ``to_ast`` method and supply the ``type`` attribute

    Expressions can be composed via overloaded arithmetic and boolean comparison operators
    to create new expressions. Note this means you cannot evaluate expression equality via
    ``==`` which produces a new expression instead of producing a python boolean.
    """

    type: ast.ClassicalType

    def to_ast(self, program: Program) -> ast.Expression:
        """Converts the oqpy expression into an ast node."""
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def _to_binary(
        op_name: str,
        first: AstConvertible,
        second: AstConvertible,
        result_type: ast.ClassicalType | None = None,
    ) -> OQPyBinaryExpression:
        """Helper method to produce a binary expression."""
        return OQPyBinaryExpression(ast.BinaryOperator[op_name], first, second, result_type)

    @staticmethod
    def _to_unary(op_name: str, exp: AstConvertible) -> OQPyUnaryExpression:
        """Helper method to produce a binary expression."""
        return OQPyUnaryExpression(ast.UnaryOperator[op_name], exp)

    def __pos__(self) -> OQPyExpression:
        return self

    def __neg__(self) -> OQPyUnaryExpression:
        return self._to_unary("-", self)

    def __add__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("+", self, other)

    def __radd__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("+", other, self)

    def __sub__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("-", self, other)

    def __rsub__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("-", other, self)

    def __mod__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("%", self, other)

    def __rmod__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("%", other, self)

    def __mul__(self, other: AstConvertible) -> OQPyBinaryExpression:
        result_type = compute_product_types(self, other)
        return self._to_binary("*", self, other, result_type)

    def __rmul__(self, other: AstConvertible) -> OQPyBinaryExpression:
        result_type = compute_product_types(other, self)
        return self._to_binary("*", other, self, result_type)

    def __truediv__(self, other: AstConvertible) -> OQPyBinaryExpression:
        result_type = compute_quotient_types(self, other)
        return self._to_binary("/", self, other, result_type)

    def __rtruediv__(self, other: AstConvertible) -> OQPyBinaryExpression:
        result_type = compute_quotient_types(other, self)
        return self._to_binary("/", other, self, result_type)

    def __pow__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("**", self, other)

    def __rpow__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("**", other, self)

    def __lshift__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("<<", self, other)

    def __rlshift__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("<<", other, self)

    def __rshift__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary(">>", self, other)

    def __rrshift__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary(">>", other, self)

    def __and__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("&", self, other)

    def __rand__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("&", other, self)

    def __or__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("|", self, other)

    def __ror__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("|", other, self)

    def __xor__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("^", self, other)

    def __rxor__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("^", other, self)

    def __eq__(self, other: AstConvertible) -> OQPyBinaryExpression:  # type: ignore[override]
        return self._to_binary("==", self, other)

    def __ne__(self, other: AstConvertible) -> OQPyBinaryExpression:  # type: ignore[override]
        return self._to_binary("!=", self, other)

    def __gt__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary(">", self, other)

    def __lt__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("<", self, other)

    def __ge__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary(">=", self, other)

    def __le__(self, other: AstConvertible) -> OQPyBinaryExpression:
        return self._to_binary("<=", self, other)

    def __invert__(self) -> OQPyUnaryExpression:
        return self._to_unary("~", self)

    def __bool__(self) -> bool:
        raise RuntimeError(
            "OQPy expressions cannot be converted to bool. This can occur if you try to check "
            "the equality of expressions using == instead of expr_matches."
        )


def _get_type(val: AstConvertible) -> ast.ClassicalType:
    if isinstance(val, OQPyExpression):
        return val.type
    elif isinstance(val, int):
        return ast.IntType()
    elif isinstance(val, float):
        return ast.FloatType()
    elif isinstance(val, complex):
        return ast.ComplexType(ast.FloatType())
    else:
        raise ValueError(f"Cannot multiply/divide oqpy expression with with {type(val)}")


def compute_product_types(left: AstConvertible, right: AstConvertible) -> ast.ClassicalType:
    """Find the result type for a product of two terms."""
    left_type = _get_type(left)
    right_type = _get_type(right)

    types_map = {
        (ast.FloatType, ast.FloatType): left_type,
        (ast.FloatType, ast.IntType): left_type,
        (ast.FloatType, ast.UintType): left_type,
        (ast.FloatType, ast.DurationType): right_type,
        (ast.FloatType, ast.AngleType): right_type,
        (ast.FloatType, ast.ComplexType): right_type,
        (ast.IntType, ast.FloatType): right_type,
        (ast.IntType, ast.IntType): left_type,
        (ast.IntType, ast.UintType): left_type,
        (ast.IntType, ast.DurationType): right_type,
        (ast.IntType, ast.AngleType): right_type,
        (ast.IntType, ast.ComplexType): right_type,
        (ast.UintType, ast.FloatType): right_type,
        (ast.UintType, ast.IntType): right_type,
        (ast.UintType, ast.UintType): left_type,
        (ast.UintType, ast.DurationType): right_type,
        (ast.UintType, ast.AngleType): right_type,
        (ast.UintType, ast.ComplexType): right_type,
        (ast.DurationType, ast.FloatType): left_type,
        (ast.DurationType, ast.IntType): left_type,
        (ast.DurationType, ast.UintType): left_type,
        (ast.DurationType, ast.DurationType): TypeError(
            "Cannot multiply two durations. You may need to re-group computations to eliminate this."
        ),
        (ast.DurationType, ast.AngleType): TypeError("Cannot multiply duration and angle"),
        (ast.DurationType, ast.ComplexType): TypeError("Cannot multiply duration and complex"),
        (ast.AngleType, ast.FloatType): left_type,
        (ast.AngleType, ast.IntType): left_type,
        (ast.AngleType, ast.UintType): left_type,
        (ast.AngleType, ast.DurationType): TypeError("Cannot multiply angle and duration"),
        (ast.AngleType, ast.AngleType): TypeError("Cannot multiply two angles"),
        (ast.AngleType, ast.ComplexType): TypeError("Cannot multiply angle and complex"),
        (ast.ComplexType, ast.FloatType): left_type,
        (ast.ComplexType, ast.IntType): left_type,
        (ast.ComplexType, ast.UintType): left_type,
        (ast.ComplexType, ast.DurationType): TypeError("Cannot multiply complex and duration"),
        (ast.ComplexType, ast.AngleType): TypeError("Cannot multiply complex and angle"),
        (ast.ComplexType, ast.ComplexType): left_type,
    }

    try:
        result_type = types_map[type(left_type), type(right_type)]
    except KeyError as e:
        raise TypeError(f"Could not identify types for product {left} and {right}") from e
    if isinstance(result_type, Exception):
        raise result_type
    return result_type


def compute_quotient_types(left: AstConvertible, right: AstConvertible) -> ast.ClassicalType:
    """Find the result type for a quotient of two terms."""
    left_type = _get_type(left)
    right_type = _get_type(right)
    float_type = ast.FloatType()

    types_map = {
        (ast.FloatType, ast.FloatType): left_type,
        (ast.FloatType, ast.IntType): left_type,
        (ast.FloatType, ast.UintType): left_type,
        (ast.FloatType, ast.DurationType): TypeError("Cannot divide float by duration"),
        (ast.FloatType, ast.AngleType): TypeError("Cannot divide float by angle"),
        (ast.FloatType, ast.ComplexType): right_type,
        (ast.IntType, ast.FloatType): right_type,
        (ast.IntType, ast.IntType): float_type,
        (ast.IntType, ast.UintType): float_type,
        (ast.IntType, ast.DurationType): TypeError("Cannot divide int by duration"),
        (ast.IntType, ast.AngleType): TypeError("Cannot divide int by angle"),
        (ast.IntType, ast.ComplexType): right_type,
        (ast.UintType, ast.FloatType): right_type,
        (ast.UintType, ast.IntType): float_type,
        (ast.UintType, ast.UintType): float_type,
        (ast.UintType, ast.DurationType): TypeError("Cannot divide uint by duration"),
        (ast.UintType, ast.AngleType): TypeError("Cannot divide uint by angle"),
        (ast.UintType, ast.ComplexType): right_type,
        (ast.DurationType, ast.FloatType): left_type,
        (ast.DurationType, ast.IntType): left_type,
        (ast.DurationType, ast.UintType): left_type,
        (ast.DurationType, ast.DurationType): ast.FloatType(),
        (ast.DurationType, ast.AngleType): TypeError("Cannot divide duration by angle"),
        (ast.DurationType, ast.ComplexType): TypeError("Cannot divide duration by complex"),
        (ast.AngleType, ast.FloatType): left_type,
        (ast.AngleType, ast.IntType): left_type,
        (ast.AngleType, ast.UintType): left_type,
        (ast.AngleType, ast.DurationType): TypeError("Cannot divide by duration"),
        (ast.AngleType, ast.AngleType): float_type,
        (ast.AngleType, ast.ComplexType): TypeError("Cannot divide by angle by complex"),
        (ast.ComplexType, ast.FloatType): left_type,
        (ast.ComplexType, ast.IntType): left_type,
        (ast.ComplexType, ast.UintType): left_type,
        (ast.ComplexType, ast.DurationType): TypeError("Cannot divide by duration"),
        (ast.ComplexType, ast.AngleType): TypeError("Cannot divide by angle"),
        (ast.ComplexType, ast.ComplexType): left_type,
    }

    try:
        result_type = types_map[type(left_type), type(right_type)]
    except KeyError as e:
        raise TypeError(f"Could not identify types for quotient {left} and {right}") from e
    if isinstance(result_type, Exception):
        raise result_type
    return result_type


def logical_and(first: AstConvertible, second: AstConvertible) -> OQPyBinaryExpression:
    """Logical AND."""
    return OQPyBinaryExpression(ast.BinaryOperator["&&"], first, second)


def logical_or(first: AstConvertible, second: AstConvertible) -> OQPyBinaryExpression:
    """Logical OR."""
    return OQPyBinaryExpression(ast.BinaryOperator["||"], first, second)


def expr_matches(a: Any, b: Any) -> bool:
    """Check equality of the given objects.

    This bypasses calling ``__eq__`` on expr objects.
    """
    if type(a) is not type(b):
        return False
    if isinstance(a, (list, np.ndarray)):
        if len(a) != len(b):
            return False
        return all(expr_matches(ai, bi) for ai, bi in zip(a, b))
    elif isinstance(a, dict):
        if a.keys() != b.keys():
            return False
        return all(expr_matches(va, b[k]) for k, va in a.items())
    if hasattr(a, "__dict__"):
        return expr_matches(a.__dict__, b.__dict__)
    else:
        return a == b


@runtime_checkable
class ExpressionConvertible(Protocol):
    """This is the protocol an object can implement in order to be usable as an expression."""

    def _to_oqpy_expression(self) -> HasToAst:
        ...


class OQPyUnaryExpression(OQPyExpression):
    """An expression consisting of one expression preceded by an operator."""

    def __init__(self, op: ast.UnaryOperator, exp: AstConvertible):
        super().__init__()
        self.op = op
        self.exp = exp
        if isinstance(exp, OQPyExpression):
            self.type = exp.type
        else:
            raise TypeError("exp is an expression")

    def to_ast(self, program: Program) -> ast.UnaryExpression:
        """Converts the OQpy expression into an ast node."""
        return ast.UnaryExpression(self.op, to_ast(program, self.exp))


class OQPyBinaryExpression(OQPyExpression):
    """An expression consisting of two subexpressions joined by an operator."""

    def __init__(
        self,
        op: ast.BinaryOperator,
        lhs: AstConvertible,
        rhs: AstConvertible,
        ast_type: ast.ClassicalType | None = None,
    ):
        super().__init__()
        self.op = op
        self.lhs = lhs
        self.rhs = rhs
        # TODO (#9): More robust type checking which considers both arguments
        #   types, as well as the operator.
        if ast_type is None:
            if isinstance(lhs, OQPyExpression):
                ast_type = lhs.type
            elif isinstance(rhs, OQPyExpression):
                ast_type = rhs.type
            else:
                raise TypeError("Neither lhs nor rhs is an expression?")
        self.type = ast_type

        # Adding floats to durations is not allowed. So we promote types as necessary.
        if isinstance(self.type, ast.DurationType) and self.op in [
            ast.BinaryOperator["+"],
            ast.BinaryOperator["-"],
        ]:
            # Late import to avoid circular imports.
            from oqpy.timing import convert_float_to_duration

            self.lhs = convert_float_to_duration(self.lhs)
            self.rhs = convert_float_to_duration(self.rhs)

    def to_ast(self, program: Program) -> ast.BinaryExpression:
        """Converts the OQpy expression into an ast node."""
        return ast.BinaryExpression(self.op, to_ast(program, self.lhs), to_ast(program, self.rhs))


class Var(ABC):
    """Abstract base class for both classical and quantum variables."""

    def __init__(self, name: str, needs_declaration: bool = True):
        self.name = name
        self._needs_declaration = needs_declaration

    def _var_matches(self, other: Any) -> bool:
        """Return true if this object represents the same variable as other.

        Needed because we overload ``==`` for expressions.
        """
        if isinstance(self, OQPyExpression):
            return expr_matches(self, other)
        else:
            return self == other

    @abstractmethod
    def to_ast(self, program: Program) -> ast.Expression:
        """Converts the OQpy variable into an ast node."""
        ...

    @abstractmethod
    def make_declaration_statement(self, program: Program) -> ast.Statement:
        """Make an ast statement that declares the OQpy variable."""
        ...


@runtime_checkable
class HasToAst(Protocol):
    """Protocol for objects which can be converted into ast nodes."""

    def to_ast(self, program: Program) -> ast.Expression:
        """Converts the OQpy object into an ast node."""
        ...


AstConvertible = Union[
    HasToAst, bool, int, float, complex, Iterable, ExpressionConvertible, ast.Expression
]


def to_ast(program: Program, item: AstConvertible) -> ast.Expression:
    """Convert an object to an AST node."""
    if hasattr(item, "_to_oqpy_expression"):
        return item._to_oqpy_expression().to_ast(program)  # type: ignore[union-attr]
    if isinstance(item, (complex, np.complexfloating)):
        if item.imag == 0:
            return to_ast(program, item.real)
        if item.real == 0:
            if item.imag < 0:
                return ast.UnaryExpression(ast.UnaryOperator["-"], ast.ImaginaryLiteral(-item.imag))
            else:
                return ast.ImaginaryLiteral(item.imag)
        if item.imag < 0:
            return ast.BinaryExpression(
                ast.BinaryOperator["-"],
                ast.FloatLiteral(item.real),
                ast.ImaginaryLiteral(-item.imag),
            )
        return ast.BinaryExpression(
            ast.BinaryOperator["+"], ast.FloatLiteral(item.real), ast.ImaginaryLiteral(item.imag)
        )
    if isinstance(item, (bool, np.bool_)):
        return ast.BooleanLiteral(item)
    if isinstance(item, (int, np.integer)):
        item = int(item)
        if item < 0:
            return ast.UnaryExpression(ast.UnaryOperator["-"], ast.IntegerLiteral(-item))
        return ast.IntegerLiteral(item)
    if isinstance(item, (float, np.floating)):
        if item < 0:
            if program.simplify_constants:
                neg_ast_term = detect_and_convert_constants(-item, program)
            else:
                neg_ast_term = ast.FloatLiteral(-item)
            return ast.UnaryExpression(ast.UnaryOperator["-"], neg_ast_term)
        if program.simplify_constants:
            return detect_and_convert_constants(item, program)
        return ast.FloatLiteral(item)
    if isinstance(item, Iterable):
        return ast.ArrayLiteral([to_ast(program, i) for i in item])
    if isinstance(item, ast.Expression):
        return item
    if hasattr(item, "to_ast"):  # Using isinstance(HasToAst) slowish
        return item.to_ast(program)  # type: ignore[union-attr]
    raise TypeError(f"Cannot convert {item} of type {type(item)} to ast")


def optional_ast(program: Program, item: AstConvertible | None) -> ast.Expression | None:
    """Convert item to ast if it is not None."""
    if item is None:
        return None
    return to_ast(program, item)


def map_to_ast(program: Program, items: Iterable[AstConvertible]) -> list[ast.Expression]:
    """Convert a sequence of items into a sequence of ast nodes."""
    return [to_ast(program, item) for item in items]


def make_annotations(vals: Sequence[str | tuple[str, str]]) -> list[ast.Annotation]:
    """Convert strings/tuples of strings into Annotation ast nodes."""
    anns: list[ast.Annotation] = []
    for val in vals:
        if isinstance(val, str):
            anns.append(ast.Annotation(val))
        else:
            keyword, command = val
            anns.append(ast.Annotation(keyword, command))
    return anns


def detect_and_convert_constants(val: float | np.floating[Any], program: Program) -> ast.Expression:
    """Construct a float ast expression which is either a literal or an expression using constants."""
    if val == 0:
        return ast.FloatLiteral(val)
    if val < 0.5 or val > 100:
        return ast.FloatLiteral(val)
    x = val / (math.pi / 4.0)
    rx = round(x)
    if not math.isclose(x, rx, rel_tol=1e-12):
        return ast.FloatLiteral(val)
    term: OQPyExpression
    if rx == 4:
        term = classical_types.pi
    elif rx == 2:
        term = classical_types.pi / 2
    elif rx == 1:
        term = classical_types.pi / 4
    elif rx % 4 == 0:
        term = (rx // 4) * classical_types.pi
    elif rx % 2 == 0:
        term = (rx // 2) * classical_types.pi / 2
    else:
        term = rx * classical_types.pi / 4
    return term.to_ast(program)



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Contains methods for producing subroutines and extern function calls."""

from __future__ import annotations

import functools
import inspect
from typing import Any, Callable, Sequence, TypeVar, get_type_hints

from mypy_extensions import VarArg
from openpulse import ast

import oqpy.program
from oqpy.base import AstConvertible, OQPyExpression, make_annotations, to_ast
from oqpy.classical_types import OQFunctionCall, _ClassicalVar
from oqpy.quantum_types import Qubit
from oqpy.timing import convert_float_to_duration

__all__ = ["subroutine", "declare_extern", "declare_waveform_generator"]

SubroutineParams = [oqpy.Program, VarArg(AstConvertible)]

FnType = TypeVar("FnType", bound=Callable[..., Any])


def enable_decorator_arguments(f: FnType) -> Callable[..., FnType]:
    @functools.wraps(f)
    def decorator(*args, **kwargs):  # type: ignore[no-untyped-def]
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return f(args[0])
        else:
            return lambda realf: f(realf, *args, **kwargs)

    return decorator


@enable_decorator_arguments
def subroutine(
    func: Callable[[oqpy.Program, VarArg(AstConvertible)], AstConvertible | None],
    annotations: Sequence[str | tuple[str, str]] = (),
) -> Callable[[oqpy.Program, VarArg(AstConvertible)], OQFunctionCall]:
    """Decorator to declare a subroutine.

    The function should take a program as well as any other arguments required.
    Note that the decorated function must include type hints for all arguments,
    and (other than the initial program) all of these type hints must be oqpy
    Variable types.

    .. code-block:: python

        @subroutine(annotations=("optimize", "-O3"))
        def increment_variable(program: Program, i: IntVar):
            program.increment(i, 1)

        j = IntVar(0)
        increment_variable(j)

    This should generate the following OpenQASM:

    .. code-block:: qasm3

        @optimize -O3
        def increment_variable(int[32] i) {
            i += 1;
        }

        int[32] j = 0;
        increment_variable(j);

    Args:
        func (Callable[[oqpy.Program, VarArg(AstConvertible)], AstConvertible | None]):
            function to decorate. Its first argument must be an OQpy program.
        annotations (Sequence[str | tuple[str, str]]): a collection of strings or
            tuples of string that annotate the subroutine.

    Returns:
        Callable[[oqpy.Program, VarArg(AstConvertible)], AstConvertible | None]:
            decorated function with added subroutine_declaration attribute.
    """
    name = func.__name__
    identifier = ast.Identifier(func.__name__)
    argnames = list(inspect.signature(func).parameters.keys())
    type_hints = get_type_hints(func)
    inputs = {}  # used as inputs when calling the actual python function
    arguments = []  # used in the ast definition of the subroutine
    for argname in argnames[1:]:  # arg 0 should be program
        if argname not in type_hints:
            raise ValueError(f"No type hint provided for {argname} on subroutine {name}.")
        input_ = inputs[argname] = type_hints[argname](name=argname)

        if isinstance(input_, _ClassicalVar):
            arguments.append(ast.ClassicalArgument(input_.type, ast.Identifier(argname)))
        elif isinstance(input_, Qubit):
            arguments.append(ast.QuantumArgument(ast.Identifier(input_.name), None))
        else:
            raise ValueError(
                f"Type hint for {argname} on subroutine {name} is not an oqpy variable type."
            )

    inner_prog = oqpy.Program()
    for input_val in inputs.values():
        inner_prog._mark_var_declared(input_val)
    output = func(inner_prog, **inputs)
    inner_prog.autodeclare()
    inner_prog._state.finalize_if_clause()
    body = inner_prog._state.body
    if isinstance(output, OQPyExpression):
        return_type = output.type
        body.append(ast.ReturnStatement(to_ast(inner_prog, output)))
    elif output is None:
        return_type = None
        if type_hints.get("return", False):
            return_hint = type_hints["return"]()
            if isinstance(return_hint, _ClassicalVar):
                return_type = return_hint.type
            elif return_hint is not None:
                raise ValueError(
                    f"Type hint for return variable on subroutine {name} is not an oqpy classical type."
                )
    else:
        raise ValueError("Output type of subroutine {name} was neither oqpy expression nor None.")
    stmt = ast.SubroutineDefinition(
        identifier,
        arguments=arguments,
        return_type=return_type,
        body=body,
    )
    stmt.annotations = make_annotations(annotations)

    @functools.wraps(func)
    def wrapper(
        program: oqpy.Program,
        *args: AstConvertible,
    ) -> OQFunctionCall:
        program.defcals.update(inner_prog.defcals)
        for name, subroutine_stmt in inner_prog.subroutines.items():
            program._add_subroutine(name, subroutine_stmt)
        program.externs.update(inner_prog.externs)
        return OQFunctionCall(
            identifier,
            args,
            return_type,
            subroutine_decl=stmt,
        )

    setattr(wrapper, "subroutine_declaration", (name, stmt))
    return wrapper


def declare_extern(
    name: str,
    args: list[tuple[str, ast.ClassicalType]],
    return_type: ast.ClassicalType,
    annotations: Sequence[str | tuple[str, str]] = (),
) -> Callable[..., OQFunctionCall]:
    """Declare an extern and return a callable which adds the extern.

    .. code-block:: python

        sqrt = declare_extern(
            "sqrt",
            [("x", classical_types.float32)],
            classical_types.float32,
        )
        var = FloatVar[32]()
        program.set(var, sqrt(0.5))

    """
    arg_names = list(zip(*(args)))[0]
    arg_types = list(zip(*(args)))[1]
    extern_decl = ast.ExternDeclaration(
        ast.Identifier(name),
        [ast.ExternArgument(type=t) for t in arg_types],
        return_type,
    )
    extern_decl.annotations = make_annotations(annotations)

    def call_extern(*call_args: AstConvertible, **call_kwargs: AstConvertible) -> OQFunctionCall:
        new_args = list(call_args) + [None] * len(call_kwargs)

        # Testing that the number of arguments is equal to what's defined by the prototype
        if len(new_args) != len(args):
            raise TypeError(
                f"{name}() takes {len(args)} positional arguments but {len(new_args)} were given."
            )

        # Adding keyword arguments to the list of arguments
        for k in call_kwargs:
            try:
                k_idx = arg_names.index(k)
            except ValueError:
                raise TypeError(f"{name}() got an unexpected keyword argument '{k}'.")

            if k_idx < len(call_args):
                raise TypeError(f"{name}() got multiple values for argument '{k}'.")

            if type(arg_types[k_idx]) == ast.DurationType:
                new_args[k_idx] = convert_float_to_duration(call_kwargs[k])
            else:
                new_args[k_idx] = call_kwargs[k]

        # Casting floats into durations for the non-keyword arguments
        for i, a in enumerate(call_args):
            if type(arg_types[i]) == ast.DurationType:
                new_args[i] = convert_float_to_duration(a)
        return OQFunctionCall(name, new_args, return_type, extern_decl=extern_decl)

    return call_extern


def declare_waveform_generator(
    name: str,
    argtypes: list[tuple[str, ast.ClassicalType]],
    annotations: Sequence[str | tuple[str, str]] = (),
) -> Callable[..., OQFunctionCall]:
    """Create a function which generates waveforms using a specified name and argument signature."""
    func = declare_extern(name, argtypes, ast.WaveformType(), annotations=annotations)
    return func



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Classes representing oqpy variables with classical types."""

from __future__ import annotations

import functools
import random
import string
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from openpulse import ast

from oqpy.base import (
    AstConvertible,
    OQPyExpression,
    Var,
    make_annotations,
    map_to_ast,
    optional_ast,
    to_ast,
)
from oqpy.timing import convert_float_to_duration

if TYPE_CHECKING:
    from typing import Literal

    from oqpy.program import Program

__all__ = [
    "pi",
    "ArrayVar",
    "BoolVar",
    "IntVar",
    "UintVar",
    "FloatVar",
    "AngleVar",
    "BitVar",
    "ComplexVar",
    "DurationVar",
    "OQFunctionCall",
    "OQIndexExpression",
    "StretchVar",
    "_ClassicalVar",
    "duration",
    "stretch",
    "bool_",
    "bit_",
    "bit",
    "bit8",
    "convert_range",
    "int_",
    "int32",
    "int64",
    "uint_",
    "uint32",
    "uint64",
    "float_",
    "float32",
    "float64",
    "complex_",
    "complex64",
    "complex128",
    "angle_",
    "angle32",
]

# The following methods and constants are useful for creating signatures
# for openqasm function calls, as is required when specifying
# waveform generating methods.
# If you wish to create a variable with a particular type, please use the
# subclasses of ``_ClassicalVar`` instead.


def int_(size: int | None = None) -> ast.IntType:
    """Create a sized signed integer type."""
    return ast.IntType(ast.IntegerLiteral(size) if size is not None else None)


def uint_(size: int | None = None) -> ast.UintType:
    """Create a sized unsigned integer type."""
    return ast.UintType(ast.IntegerLiteral(size) if size is not None else None)


def float_(size: int | None = None) -> ast.FloatType:
    """Create a sized floating-point type."""
    return ast.FloatType(ast.IntegerLiteral(size) if size is not None else None)


def angle_(size: int | None = None) -> ast.AngleType:
    """Create a sized angle type."""
    return ast.AngleType(ast.IntegerLiteral(size) if size is not None else None)


def complex_(size: int) -> ast.ComplexType:
    """Create a sized complex type.

    Note the size represents the total size, and thus the components have
    half of the requested size.
    """
    return ast.ComplexType(ast.FloatType(ast.IntegerLiteral(size // 2)))


def bit_(size: int | None = None) -> ast.BitType:
    """Create a sized bit type."""
    return ast.BitType(ast.IntegerLiteral(size) if size is not None else None)


duration = ast.DurationType()
stretch = ast.StretchType()
bool_ = ast.BoolType()
bit = ast.BitType()
bit8 = bit_(8)
int32 = int_(32)
int64 = int_(64)
uint32 = uint_(32)
uint64 = uint_(64)
float32 = float_(32)
float64 = float_(64)
complex64 = complex_(64)
complex128 = complex_(128)
angle32 = angle_(32)


def convert_range(program: Program, item: Union[slice, range]) -> ast.RangeDefinition:
    """Convert a slice or range into an ast node."""
    return ast.RangeDefinition(
        to_ast(program, item.start),
        to_ast(program, item.stop - 1),
        to_ast(program, item.step) if item.step != 1 else None,
    )


class Identifier(OQPyExpression):
    """Base class to specify constant symbols."""

    name: str

    def __init__(self, name: str, ast_type: ast.ClassicalType) -> None:
        self.name = name
        self.type = ast_type

    def to_ast(self, program: Program) -> ast.Expression:
        return ast.Identifier(name=self.name)


pi = Identifier(name="pi", ast_type=ast.FloatType())


class _ClassicalVar(Var, OQPyExpression):
    """Base type for variables with classical type.

    Subclasses should supply the type_cls class variable.
    """

    type_cls: Type[ast.ClassicalType]

    def __init__(
        self,
        init_expression: AstConvertible | Literal["input", "output"] | None = None,
        name: str | None = None,
        needs_declaration: bool = True,
        annotations: Sequence[str | tuple[str, str]] = (),
        **type_kwargs: Any,
    ):
        name = name or "".join([random.choice(string.ascii_letters) for _ in range(10)])
        super().__init__(name, needs_declaration=needs_declaration)
        self.type = self.type_cls(**type_kwargs)
        self.init_expression = init_expression
        self.annotations = annotations

    def to_ast(self, program: Program) -> ast.Identifier:
        """Converts the OQpy variable into an ast node."""
        program._add_var(self)
        return ast.Identifier(self.name)

    def make_declaration_statement(self, program: Program) -> ast.Statement:
        """Make an ast statement that declares the OQpy variable."""
        if isinstance(self.init_expression, str) and self.init_expression in ("input", "output"):
            return ast.IODeclaration(
                ast.IOKeyword[self.init_expression], self.type, self.to_ast(program)
            )
        init_expression_ast = optional_ast(program, self.init_expression)
        stmt = ast.ClassicalDeclaration(self.type, self.to_ast(program), init_expression_ast)
        stmt.annotations = make_annotations(self.annotations)
        return stmt


class BoolVar(_ClassicalVar):
    """An (unsized) oqpy variable with bool type."""

    type_cls = ast.BoolType


class _SizedVar(_ClassicalVar):
    """Base class for variables with a specified size."""

    default_size: int | None = None
    size: int | None

    def __class_getitem__(cls: Type[_SizedVarT], item: int) -> Callable[..., _SizedVarT]:
        # Allows IntVar[64]() notation
        return functools.partial(cls, size=item)

    def __init__(self, *args: Any, size: int | None = None, **kwargs: Any):
        if size is None:
            self.size = self.default_size
        else:
            if not isinstance(size, int) or size <= 0:
                raise ValueError(
                    f"The size of '{self.type_cls}' objects must be an positive integer."
                )
            self.size = size
        super().__init__(*args, **kwargs, size=ast.IntegerLiteral(self.size) if self.size else None)


_SizedVarT = TypeVar("_SizedVarT", bound=_SizedVar)


class IntVar(_SizedVar):
    """An oqpy variable with integer type."""

    type_cls = ast.IntType
    default_size = 32


class UintVar(_SizedVar):
    """An oqpy variable with unsigned integer type."""

    type_cls = ast.UintType
    default_size = 32


class FloatVar(_SizedVar):
    """An oqpy variable with floating type."""

    type_cls = ast.FloatType
    default_size = 64


class AngleVar(_SizedVar):
    """An oqpy variable with angle type."""

    type_cls = ast.AngleType
    default_size = 32


class BitVar(_SizedVar):
    """An oqpy variable with bit type."""

    type_cls = ast.BitType

    def __getitem__(self, idx: Union[int, slice, Iterable[int]]) -> BitVar:
        if self.size is None:
            raise TypeError(f"'{self.type_cls}' object is not subscriptable")
        if isinstance(idx, int):
            if 0 <= idx < self.size:
                return BitVar(
                    init_expression=ast.IndexExpression(
                        ast.Identifier(self.name), [ast.IntegerLiteral(idx)]
                    ),
                    name=f"{self.name}[{idx}]",
                    needs_declaration=False,
                )
            else:
                raise IndexError("list index out of range.")
        else:
            raise IndexError("The list index must be an integer.")


class ComplexVar(_ClassicalVar):
    """An oqpy variable with bit type."""

    type_cls = ast.ComplexType
    base_type: ast.FloatType = float64

    def __class_getitem__(cls, item: ast.FloatType) -> Callable[..., ComplexVar]:
        return functools.partial(cls, base_type=item)

    def __init__(
        self,
        init_expression: AstConvertible | Literal["input", "output"] | None = None,
        *args: Any,
        base_type: ast.FloatType = float64,
        **kwargs: Any,
    ) -> None:
        assert isinstance(base_type, ast.FloatType)
        self.base_type = base_type

        if not isinstance(init_expression, (complex, type(None), str, OQPyExpression)):
            init_expression = complex(init_expression)  # type: ignore[arg-type]
        super().__init__(init_expression, *args, **kwargs, base_type=base_type)


class DurationVar(_ClassicalVar):
    """An oqpy variable with duration type."""

    type_cls = ast.DurationType

    def __init__(
        self,
        init_expression: AstConvertible | Literal["input", "output"] | None = None,
        name: str | None = None,
        *args: Any,
        **type_kwargs: Any,
    ) -> None:
        if init_expression is not None and not isinstance(init_expression, str):
            init_expression = convert_float_to_duration(init_expression)
        super().__init__(init_expression, name, *args, **type_kwargs)


class StretchVar(_ClassicalVar):
    """An oqpy variable with stretch type."""

    type_cls = ast.StretchType


AllowedArrayTypes = Union[_SizedVar, DurationVar, BoolVar, ComplexVar]


class ArrayVar(_ClassicalVar):
    """An oqpy array variable."""

    type_cls = ast.ArrayType
    dimensions: list[int]
    base_type: type[AllowedArrayTypes]

    def __class_getitem__(
        cls, item: tuple[type[AllowedArrayTypes], int] | type[AllowedArrayTypes]
    ) -> Callable[..., ArrayVar]:
        # Allows usage like ArrayVar[FloatVar, 32](...) or ArrayVar[FloatVar]
        if isinstance(item, tuple):
            base_type = item[0]
            dimensions = list(item[1:])
            return functools.partial(cls, dimensions=dimensions, base_type=base_type)
        else:
            return functools.partial(cls, base_type=item)

    def __init__(
        self,
        *args: Any,
        dimensions: list[int],
        base_type: type[AllowedArrayTypes] = IntVar,
        **kwargs: Any,
    ) -> None:
        self.dimensions = dimensions
        self.base_type = base_type

        # Creating a dummy variable supports IntVar[64] etc.
        base_type_instance = base_type()
        if isinstance(base_type_instance, _SizedVar):
            array_base_type = base_type_instance.type_cls(
                size=ast.IntegerLiteral(base_type_instance.size)
            )
        elif isinstance(base_type_instance, ComplexVar):
            array_base_type = base_type_instance.type_cls(base_type=base_type_instance.base_type)
        else:
            array_base_type = base_type_instance.type_cls()

        # Automatically handle Duration array.
        if base_type is DurationVar and kwargs["init_expression"] is not None:
            kwargs["init_expression"] = (
                convert_float_to_duration(i) for i in kwargs["init_expression"]
            )

        super().__init__(
            *args,
            **kwargs,
            dimensions=[ast.IntegerLiteral(dimension) for dimension in dimensions],
            base_type=array_base_type,
        )

    def __getitem__(self, index: AstConvertible) -> OQIndexExpression:
        return OQIndexExpression(collection=self, index=index)


class OQIndexExpression(OQPyExpression):
    """An oqpy expression corresponding to an index expression."""

    def __init__(self, collection: AstConvertible, index: AstConvertible):
        self.collection = collection
        self.index = index

        if isinstance(collection, ArrayVar):
            self.type = collection.base_type().type_cls()

    def to_ast(self, program: Program) -> ast.IndexExpression:
        """Converts this oqpy index expression into an ast node."""
        return ast.IndexExpression(
            collection=to_ast(program, self.collection), index=[to_ast(program, self.index)]
        )


class OQFunctionCall(OQPyExpression):
    """An oqpy expression corresponding to a function call."""

    def __init__(
        self,
        identifier: Union[str, ast.Identifier],
        args: Iterable[AstConvertible],
        return_type: ast.ClassicalType,
        extern_decl: ast.ExternDeclaration | None = None,
        subroutine_decl: ast.SubroutineDefinition | None = None,
    ):
        """Create a new OQFunctionCall instance.

        Args:
            identifier: The function name.
            args: The function arguments.
            return_type: The type returned by the function call.
            extern_decl: An optional extern declaration ast node. If present,
                this extern declaration will be added to the top of the program
                whenever this is converted to ast.
            subroutine_decl: An optional subroutine definition ast node. If present,
                this subroutine definition will be added to the top of the program
                whenever this expression is converted to ast.
        """
        super().__init__()
        if isinstance(identifier, str):
            identifier = ast.Identifier(identifier)
        self.identifier = identifier
        self.args = args
        self.type = return_type
        self.extern_decl = extern_decl
        self.subroutine_decl = subroutine_decl

    def to_ast(self, program: Program) -> ast.Expression:
        """Converts the OQpy expression into an ast node."""
        if self.extern_decl is not None:
            program.externs[self.identifier.name] = self.extern_decl
        if self.subroutine_decl is not None:
            program._add_subroutine(self.identifier.name, self.subroutine_decl)
        return ast.FunctionCall(self.identifier, map_to_ast(program, self.args))



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Constructs for manipulating sequence timing."""

from __future__ import annotations

import contextlib
import warnings
from typing import TYPE_CHECKING, Iterator, cast

from openpulse import ast

from oqpy.base import ExpressionConvertible, HasToAst, OQPyExpression, optional_ast
from oqpy.classical_types import AstConvertible

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["Box", "convert_float_to_duration", "convert_float_to_duration", "make_duration"]


@contextlib.contextmanager
def Box(program: Program, duration: AstConvertible | None = None) -> Iterator[None]:
    """Creates a section of the program with a specified duration."""
    if duration is not None:
        duration = convert_float_to_duration(duration)
    program._push()
    yield
    state = program._pop()
    program._add_statement(ast.Box(optional_ast(program, duration), state.body))


def make_duration(time: AstConvertible) -> HasToAst:
    """Make value into an expression representing a duration."""
    warnings.warn(
        "make_duration name is deprecated in favor of convert_float_to_duration",
        DeprecationWarning,
        stacklevel=2,
    )
    return convert_float_to_duration(time)


def convert_float_to_duration(time: AstConvertible) -> HasToAst:
    """Make value into an expression representing a duration."""
    if isinstance(time, (float, int)):
        return OQDurationLiteral(time)
    if isinstance(time, OQPyExpression):
        if isinstance(time.type, (ast.UintType, ast.IntType, ast.FloatType)):
            time = time * OQDurationLiteral(1)
        elif not isinstance(time.type, ast.DurationType):
            raise TypeError(f"Cannot convert expression with type {time.type} to duration")
    if hasattr(time, "to_ast"):
        return time  # type: ignore[return-value]
    if hasattr(time, "_to_oqpy_expression"):
        time = cast(ExpressionConvertible, time)
        return time._to_oqpy_expression()
    raise TypeError(
        f"Expected either float, int, HasToAst or ExpressionConverible: Got {type(time)}"
    )


def convert_duration_to_float(value: AstConvertible) -> AstConvertible:
    if isinstance(value, OQPyExpression) and isinstance(value.type, ast.DurationType):
        value = value / OQDurationLiteral(1)
    return value


class OQDurationLiteral(OQPyExpression):
    """An expression corresponding to a duration literal."""

    def __init__(self, duration_seconds: float) -> None:
        super().__init__()
        self.duration_seconds = duration_seconds
        self.type = ast.DurationType()

    def to_ast(self, program: Program) -> ast.DurationLiteral:
        # Todo (#53): make better units?
        n = program.DURATION_MAX_DIGITS
        if self.duration_seconds >= 1:
            return ast.DurationLiteral(round(self.duration_seconds, n), ast.TimeUnit.s)
        if self.duration_seconds >= 1e-3:
            return ast.DurationLiteral(round(1e3 * self.duration_seconds, n), ast.TimeUnit.ms)
        if self.duration_seconds >= 1e-6:
            return ast.DurationLiteral(round(1e6 * self.duration_seconds, n), ast.TimeUnit.us)
        return ast.DurationLiteral(round(1e9 * self.duration_seconds, n), ast.TimeUnit.ns)



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Context manager objects used for creating control flow contexts."""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Iterable, Iterator, Optional, TypeVar, overload

from openpulse import ast

from oqpy.base import OQPyExpression, to_ast
from oqpy.classical_types import (
    AstConvertible,
    DurationVar,
    IntVar,
    _ClassicalVar,
    convert_range,
)
from oqpy.timing import convert_float_to_duration

ClassicalVarT = TypeVar("ClassicalVarT", bound=_ClassicalVar)

if TYPE_CHECKING:
    from oqpy.program import Program


__all__ = ["If", "Else", "ForIn", "While", "Range"]


@contextlib.contextmanager
def If(program: Program, condition: OQPyExpression) -> Iterator[None]:
    """Context manager for doing conditional evaluation.

    .. code-block:: python

        i = IntVar(1)
        with If(program, i == 1):
            program.increment(i)

    """
    program._push()
    yield
    state = program._pop()
    program._state.add_if_clause(to_ast(program, condition), state.body)


@contextlib.contextmanager
def Else(program: Program) -> Iterator[None]:
    """Context manager for conditional evaluation. Must come after an "If" context block.

    .. code-block:: python

        i = IntVar(1)
        with If(program, i == 1):
            program.increment(i, 1)
        with Else(program):
            program.increment(i, 2)

    """
    program._push()
    yield
    state = program._pop()
    program._state.add_else_clause(state.body)


# Overloads needed due mypy bug, see
# github.com/python/mypy/issues/8739
# https://github.com/python/mypy/issues/3737
@overload
def ForIn(
    program: Program,
    iterator: Iterable[AstConvertible] | range | AstConvertible,
    identifier_name: Optional[str],
) -> contextlib._GeneratorContextManager[IntVar]:
    ...


@overload
def ForIn(
    program: Program,
    iterator: Iterable[AstConvertible] | range | AstConvertible,
    identifier_name: Optional[str],
    identifier_type: type[ClassicalVarT],
) -> contextlib._GeneratorContextManager[ClassicalVarT]:
    ...


@contextlib.contextmanager
def ForIn(
    program: Program,
    iterator: Iterable[AstConvertible] | range | AstConvertible,
    identifier_name: Optional[str] = None,
    identifier_type: type[ClassicalVarT] | type[IntVar] = IntVar,
) -> Iterator[ClassicalVarT | IntVar]:
    """Context manager for looping a particular portion of a program.

    .. code-block:: python

        i = IntVar(1)
        with ForIn(program, range(1, 10)) as index:
            program.increment(i, index)

    """
    program._push()
    var = identifier_type(name=identifier_name, needs_declaration=False)
    yield var
    state = program._pop()

    if isinstance(iterator, range):
        # A range can only be iterated over integers.
        assert identifier_type is IntVar, "A range can only be looped over an integer."
        set_declaration = convert_range(program, iterator)
    elif isinstance(iterator, Iterable):
        if identifier_type is DurationVar:
            iterator = (convert_float_to_duration(i) for i in iterator)

        set_declaration = ast.DiscreteSet([to_ast(program, i) for i in iterator])
    else:
        set_declaration = to_ast(program, iterator)

    stmt = ast.ForInLoop(
        identifier_type.type_cls(), var.to_ast(program), set_declaration, state.body
    )
    program._add_statement(stmt)


class Range:
    """AstConvertible which creates an integer range.

    Unlike builtin python range, this allows the components to be AstConvertible,
    instead of just int.
    """

    def __init__(self, start: AstConvertible, stop: AstConvertible, step: AstConvertible = 1):
        self.start = start
        self.stop = stop
        self.step = step

    def to_ast(self, program: Program) -> ast.Expression:
        """Convert to an ast.RangeDefinition."""
        return ast.RangeDefinition(
            to_ast(program, self.start),
            ast.BinaryExpression(
                lhs=to_ast(program, self.stop),
                op=ast.BinaryOperator["-"],
                rhs=ast.IntegerLiteral(value=1),
            ),
            to_ast(program, self.step) if self.step != 1 else None,
        )


@contextlib.contextmanager
def While(program: Program, condition: OQPyExpression) -> Iterator[None]:
    """Context manager for looping a repeating a portion of a program while a condition is True.

    .. code-block:: python

        i = IntVar(1)
        with While(program, i < 5) as index:
            program.increment(i, 1)

    """
    program._push()
    yield
    state = program._pop()
    program._add_statement(ast.WhileLoop(to_ast(program, condition), state.body))



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Classes representing OpenPulse variable types."""

from __future__ import annotations

from typing import Any, Sequence

from openpulse import ast

from oqpy.base import AstConvertible
from oqpy.classical_types import OQFunctionCall, _ClassicalVar

__all__ = ["PortVar", "WaveformVar", "FrameVar", "port", "waveform", "frame"]

port = ast.PortType()
waveform = ast.WaveformType()
frame = ast.FrameType()


class PortVar(_ClassicalVar):
    """A variable type corresponding to an OpenPulse port."""

    type_cls = ast.PortType

    def __init__(self, name: str | None = None, **kwargs: Any):
        super().__init__(name=name, **kwargs)


class WaveformVar(_ClassicalVar):
    """A variable type corresponding to an OpenPulse waveform."""

    type_cls = ast.WaveformType

    def __init__(
        self, init_expression: AstConvertible | None = None, name: str | None = None, **kwargs: Any
    ):
        super().__init__(
            init_expression=init_expression,
            name=name,
            **kwargs,
        )


class FrameVar(_ClassicalVar):
    """A variable type corresponding to an OpenPulse frame."""

    type_cls = ast.FrameType

    def __init__(
        self,
        port: PortVar | None = None,
        frequency: AstConvertible | None = None,
        phase: AstConvertible = 0,
        name: str | None = None,
        needs_declaration: bool = True,
        annotations: Sequence[str | tuple[str, str]] = (),
    ):
        if (port is None) != (frequency is None):
            raise ValueError("Must declare both port and frequency or neither.")
        if port is None:
            init_expression = None
        else:
            assert frequency is not None
            init_expression = OQFunctionCall("newframe", [port, frequency, phase], ast.FrameType)
        super().__init__(
            init_expression, name, needs_declaration=needs_declaration, annotations=annotations
        )



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################

__version__ = "0.3.1"
from oqpy.classical_types import *
from oqpy.control_flow import *
from oqpy.program import *
from oqpy.pulse import *
from oqpy.quantum_types import *
from oqpy.subroutines import *
from oqpy.timing import *



############################################################################
#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License").
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
############################################################################
"""Module containing Program and related classes.

This module contains the oqpy.Program class, which is the primary user interface
for constructing openqasm/openpulse. The class follows the builder pattern in that
it contains a representation of the state of a program (internally represented as
AST components), and the class has methods which add to the current state.
"""

from __future__ import annotations

import warnings
from copy import deepcopy
from typing import Any, Iterable, Iterator, Optional

from openpulse import ast
from openpulse.printer import dumps
from openqasm3.visitor import QASMVisitor

from oqpy import classical_types, quantum_types
from oqpy.base import (
    AstConvertible,
    Var,
    expr_matches,
    map_to_ast,
    optional_ast,
    to_ast,
)
from oqpy.pulse import FrameVar, PortVar, WaveformVar
from oqpy.timing import convert_duration_to_float, convert_float_to_duration

__all__ = ["Program"]


class ProgramState:
    """Represents the current program state at a particular context level.

    A new ProgramState is created every time a context (such as the control
    flow constructs If/Else/ForIn/While) is created. A program will retain a
    stack of ProgramState objects for all currently open contexts.
    """

    def __init__(self) -> None:
        self.body: list[ast.Statement | ast.Pragma] = []
        self.if_clause: Optional[ast.BranchingStatement] = None
        self.annotations: list[ast.Annotation] = []

    def add_if_clause(self, condition: ast.Expression, if_clause: list[ast.Statement]) -> None:
        if_clause_annotations, self.annotations = self.annotations, []
        self.finalize_if_clause()
        self.if_clause = ast.BranchingStatement(condition, if_clause, [])
        self.if_clause.annotations = if_clause_annotations

    def add_else_clause(self, else_clause: list[ast.Statement]) -> None:
        if self.if_clause is None:
            raise RuntimeError("Else without If.")
        self.if_clause.else_block = else_clause
        self.finalize_if_clause()

    def finalize_if_clause(self) -> None:
        if self.if_clause is not None:
            if_clause, self.if_clause = self.if_clause, None
            self.add_statement(if_clause)

    def add_statement(self, stmt: ast.Statement | ast.Pragma) -> None:
        # This function accepts Statement and Pragma even though
        # it seems to conflict with the definition of ast.Program.
        # Issue raised in https://github.com/openqasm/openqasm/issues/468
        assert isinstance(stmt, (ast.Statement, ast.Pragma))
        if isinstance(stmt, ast.Statement) and self.annotations:
            stmt.annotations = self.annotations + list(stmt.annotations)
            self.annotations = []
        self.finalize_if_clause()
        self.body.append(stmt)


class Program:
    """A builder class for OpenQASM/OpenPulse programs."""

    DURATION_MAX_DIGITS = 12

    def __init__(self, version: Optional[str] = "3.0", simplify_constants: bool = True) -> None:
        self.stack: list[ProgramState] = [ProgramState()]
        self.defcals: dict[
            tuple[tuple[str, ...], str, tuple[str, ...]], ast.CalibrationDefinition
        ] = {}
        self.subroutines: dict[str, ast.SubroutineDefinition] = {}
        self.externs: dict[str, ast.ExternDeclaration] = {}
        self.declared_vars: dict[str, Var] = {}
        self.undeclared_vars: dict[str, Var] = {}
        self.simplify_constants = simplify_constants
        self.declared_subroutines: set[str] = set()

        if version is None or (
            len(version.split(".")) in [1, 2]
            and all([item.isnumeric() for item in version.split(".")])
        ):
            self.version = version
        else:
            raise RuntimeError("Version number does not match the X[.y] format.")

    def __iadd__(self, other: Program) -> Program:
        """In-place concatenation of programs."""
        if len(other.stack) > 1:
            raise RuntimeError("Cannot add subprogram with unclosed contextmanagers.")
        self._state.finalize_if_clause()
        self._state.body.extend(other._state.body)
        self._state.if_clause = other._state.if_clause
        self._state.finalize_if_clause()
        self.defcals.update(other.defcals)
        for name, subroutine_stmt in other.subroutines.items():
            self._add_subroutine(
                name, subroutine_stmt, needs_declaration=name not in other.declared_subroutines
            )
        self.externs.update(other.externs)
        for var in other.declared_vars.values():
            self._mark_var_declared(var)
        for var in other.undeclared_vars.values():
            self._add_var(var)
        return self

    def __add__(self, other: Program) -> Program:
        """Return concatenation of two programs."""
        assert isinstance(other, Program)
        self_copy = deepcopy(self)
        self_copy += other
        return self_copy

    @property
    def _state(self) -> ProgramState:
        """The current program state is found on the top of the stack."""
        return self.stack[-1]

    @property
    def frame_vars(self) -> Iterator[FrameVar]:
        """Returns an Iterator of any declared/undeclared FrameVar used in the program."""
        for v in {**self.declared_vars, **self.undeclared_vars}.values():
            if isinstance(v, FrameVar):
                yield v

    @property
    def waveform_vars(self) -> Iterator[WaveformVar]:
        """Returns an Iterator of any declared/undeclared WaveformVar used in the program."""
        for v in {**self.declared_vars, **self.undeclared_vars}.values():
            if isinstance(v, WaveformVar):
                yield v

    def _push(self) -> None:
        """Open a new context by pushing a new program state on the stack."""
        self.stack.append(ProgramState())

    def _pop(self) -> ProgramState:
        """Close a context by removing the program state from the top stack, and return it."""
        state = self.stack.pop()
        state.finalize_if_clause()
        if state.annotations:
            warnings.warn(f"Annotation(s) {state.annotations} not applied to any statement")
        return state

    def _add_var(self, var: Var) -> None:
        """Register a variable with the program.

        If the variable is not declared (and is specified to need declaration),
        the variable will be automatically declared at the top of the program
        upon conversion to ast.

        This method is safe to call on the same variable multiple times.
        """
        name = var.name
        existing_var = self.declared_vars.get(name)
        if existing_var is None:
            existing_var = self.undeclared_vars.get(name)
        if existing_var is not None and not expr_matches(var, existing_var):
            raise RuntimeError(f"Program has conflicting variables with name {name}")
        if name not in self.declared_vars:
            self.undeclared_vars[name] = var

    def _mark_var_declared(self, var: Var) -> None:
        """Indicate that a variable has been declared."""
        self._add_var(var)
        name = var.name
        new_var = self.undeclared_vars.pop(name, None)
        if new_var is not None:
            self.declared_vars[name] = new_var

    def autodeclare(self, encal: bool = False) -> None:
        """Declare any currently undeclared variables at the beginning of the program."""
        while any(v._needs_declaration for v in self.undeclared_vars.values()):
            self.declare(
                [var for var in self.undeclared_vars.values() if var._needs_declaration],
                to_beginning=True,
                encal=encal,
            )

    def _add_statement(self, stmt: ast.Statement) -> None:
        """Add a statment to the current context's program state."""
        self._state.add_statement(stmt)

    def _add_subroutine(
        self, name: str, stmt: ast.SubroutineDefinition, needs_declaration: bool = True
    ) -> None:
        """Register a subroutine which has been used.

        Subroutines are added to the top of the program upon conversion to ast.
        """
        self.subroutines[name] = stmt
        if not needs_declaration:
            self.declared_subroutines.add(name)

    def _add_defcal(
        self,
        qubit_names: list[str],
        name: str,
        arguments: list[str],
        stmt: ast.CalibrationDefinition,
    ) -> None:
        """Register a defcal defined in this program.

        Defcals are added to the top of the program upon conversion to ast.
        """
        self.defcals[(tuple(qubit_names), name, tuple(arguments))] = stmt

    def _make_externs_statements(self, auto_encal: bool = False) -> list[ast.ExternDeclaration]:
        """Return a list of extern statements for inclusion at beginning of program.

        if auto_encal is True, any externs using openpulse types will be wrapped in a Cal block.
        """
        if not auto_encal:
            return list(self.externs.values())
        openqasm_externs, openpulse_externs = [], []
        openpulse_types = ast.PortType, ast.FrameType, ast.WaveformType
        for extern_statement in self.externs.values():
            for arg in extern_statement.arguments:
                if isinstance(arg.type, openpulse_types):
                    openpulse_externs.append(extern_statement)
                    break
            else:
                if isinstance(extern_statement.return_type, openpulse_types):
                    openpulse_externs.append(extern_statement)
                else:
                    openqasm_externs.append(extern_statement)
        if openpulse_externs:
            openqasm_externs.append(ast.CalibrationStatement(body=openpulse_externs))
        return openqasm_externs

    def to_ast(
        self,
        encal: bool = False,
        include_externs: bool = True,
        ignore_needs_declaration: bool = False,
        encal_declarations: bool = False,
    ) -> ast.Program:
        """Convert to an AST program.

        Args:
            encal: If true, wrap all statements in a "Cal" block, ensuring
                that all openpulse statements are contained within a Cal block.
            include_externs: If true, for all used extern statements, include
                an extern declaration at the top of the program.
            ignore_needs_declaration: If true, the field `_needs_declaration` of
                undeclared variables is ignored and their declaration will not
                be added to the AST
            encal_declarations: If true, when declaring undeclared variables,
                if the variables have openpulse types, automatically wrap the
                declarations in cal blocks.
        """
        mutating_prog = Program(self.version, self.simplify_constants)
        mutating_prog += self

        if not ignore_needs_declaration and mutating_prog.undeclared_vars:
            mutating_prog.autodeclare(encal=encal_declarations)

        assert len(mutating_prog.stack) == 1
        mutating_prog._state.finalize_if_clause()
        if mutating_prog._state.annotations:
            warnings.warn(
                f"Annotation(s) {mutating_prog._state.annotations} not applied to any statement"
            )
        statements = []
        if include_externs:
            statements += mutating_prog._make_externs_statements(encal_declarations)
        statements += [
            mutating_prog.subroutines[subroutine_name]
            for subroutine_name in mutating_prog.subroutines
            if subroutine_name not in mutating_prog.declared_subroutines
        ] + mutating_prog._state.body
        if encal:
            statements = [ast.CalibrationStatement(statements)]
        if encal_declarations:
            statements = [ast.CalibrationGrammarDeclaration("openpulse")] + statements
        prog = ast.Program(statements=statements, version=mutating_prog.version)
        if encal_declarations:
            MergeCalStatementsPass().visit(prog)
        return prog

    def to_qasm(
        self,
        encal: bool = False,
        include_externs: bool = True,
        ignore_needs_declaration: bool = False,
        encal_declarations: bool = False,
    ) -> str:
        """Convert to QASM text.

        See to_ast for option documentation.
        """
        return dumps(
            self.to_ast(
                encal=encal,
                include_externs=include_externs,
                ignore_needs_declaration=ignore_needs_declaration,
                encal_declarations=encal_declarations,
            ),
            indent="    ",
        ).strip()

    def declare(
        self,
        variables: list[Var] | Var,
        to_beginning: bool = False,
        encal: bool = False,
    ) -> Program:
        """Declare a variable.

        Args:
            variables: A list of variables to declare, or a single variable to declare.
            to_beginning: If true, insert the declaration at the beginning of the program,
                instead of at the current point.
            encal: If true, wrap any declarations of undeclared variables with openpulse
                types in a cal block.
        """
        if isinstance(variables, (classical_types._ClassicalVar, quantum_types.Qubit)):
            variables = [variables]

        assert isinstance(variables, list)

        openpulse_vars, openqasm_vars = [], []
        for var in variables:
            if encal and isinstance(var, (PortVar, FrameVar, WaveformVar)):
                openpulse_vars.append(var)
            else:
                openqasm_vars.append(var)

        if to_beginning:
            openqasm_vars.reverse()

        for var in openqasm_vars:
            if callable(var) and hasattr(var, "subroutine_declaration"):
                name, stmt = var.subroutine_declaration  # type: ignore[attr-defined]
                self._add_subroutine(name, stmt, needs_declaration=False)
            else:
                stmt = var.make_declaration_statement(self)
                self._mark_var_declared(var)

            if to_beginning:
                self._state.body.insert(0, stmt)
            else:
                self._add_statement(stmt)

        if openpulse_vars:
            cal_stmt = ast.CalibrationStatement([])
            for var in openpulse_vars:
                stmt = var.make_declaration_statement(self)
                cal_stmt.body.append(stmt)
                self._mark_var_declared(var)
            if to_beginning:
                self._state.body.insert(0, cal_stmt)
            else:
                self._add_statement(cal_stmt)
        return self

    def delay(
        self, time: AstConvertible, qubits_or_frames: AstConvertible | Iterable[AstConvertible] = ()
    ) -> Program:
        """Apply a delay to a set of qubits or frames."""
        if not isinstance(qubits_or_frames, Iterable):
            qubits_or_frames = [qubits_or_frames]
        ast_duration = to_ast(self, convert_float_to_duration(time))
        ast_qubits_or_frames = map_to_ast(self, qubits_or_frames)
        self._add_statement(ast.DelayInstruction(ast_duration, ast_qubits_or_frames))
        return self

    def barrier(self, qubits_or_frames: Iterable[AstConvertible]) -> Program:
        """Apply a barrier to a set of qubits or frames."""
        ast_qubits_or_frames = map_to_ast(self, qubits_or_frames)
        self._add_statement(ast.QuantumBarrier(ast_qubits_or_frames))
        return self

    def function_call(self, name: str, args: Iterable[AstConvertible]) -> None:
        """Add a function call."""
        self._add_statement(
            ast.ExpressionStatement(ast.FunctionCall(ast.Identifier(name), map_to_ast(self, args)))
        )

    def play(self, frame: AstConvertible, waveform: AstConvertible) -> Program:
        """Play a waveform on a particular frame."""
        self.function_call("play", [frame, waveform])
        return self

    def capture(self, frame: AstConvertible, kernel: AstConvertible) -> Program:
        """Capture signal integrated against a kernel on a particular frame."""
        self.function_call("capture", [frame, kernel])
        return self

    def set_phase(self, frame: AstConvertible, phase: AstConvertible) -> Program:
        """Set the phase of a particular frame."""
        # We use make_float to force phase to be a unitless (i.e. non-duration) quantity.
        # Users are expected to keep track the units that are not expressible in openqasm
        # such as s^{-1}. For instance, in 2 * oqpy.pi * tppi * DurationVar(1e-8),
        # tppi is a float but has a frequency unit. This will coerce the result type
        # to a float by assuming the duration should be represented in seconds."
        self.function_call("set_phase", [frame, convert_duration_to_float(phase)])
        return self

    def shift_phase(self, frame: AstConvertible, phase: AstConvertible) -> Program:
        """Shift the phase of a particular frame."""
        self.function_call("shift_phase", [frame, convert_duration_to_float(phase)])
        return self

    def set_frequency(self, frame: AstConvertible, freq: AstConvertible) -> Program:
        """Set the frequency of a particular frame."""
        self.function_call("set_frequency", [frame, convert_duration_to_float(freq)])
        return self

    def shift_frequency(self, frame: AstConvertible, freq: AstConvertible) -> Program:
        """Shift the frequency of a particular frame."""
        self.function_call("shift_frequency", [frame, convert_duration_to_float(freq)])
        return self

    def set_scale(self, frame: AstConvertible, scale: AstConvertible) -> Program:
        """Set the amplitude scaling of a particular frame."""
        self.function_call("set_scale", [frame, convert_duration_to_float(scale)])
        return self

    def shift_scale(self, frame: AstConvertible, scale: AstConvertible) -> Program:
        """Shift the amplitude scaling of a particular frame."""
        self.function_call("shift_scale", [frame, convert_duration_to_float(scale)])
        return self

    def returns(self, expression: AstConvertible) -> Program:
        """Return a statement from a function definition or a defcal statement."""
        self._add_statement(ast.ReturnStatement(to_ast(self, expression)))
        return self

    def gate(
        self, qubits: AstConvertible | Iterable[AstConvertible], name: str, *args: Any
    ) -> Program:
        """Apply a gate to a qubit or set of qubits."""
        if isinstance(qubits, quantum_types.Qubit):
            qubits = [qubits]
        assert isinstance(qubits, Iterable)
        self._add_statement(
            ast.QuantumGate(
                [],
                ast.Identifier(name),
                map_to_ast(self, args),
                map_to_ast(self, qubits),
            )
        )
        return self

    def reset(self, qubit: quantum_types.Qubit) -> Program:
        """Reset a particular qubit."""
        self._add_statement(ast.QuantumReset(qubits=qubit.to_ast(self)))
        return self

    def measure(
        self, qubit: quantum_types.Qubit, output_location: classical_types.BitVar | None = None
    ) -> Program:
        """Measure a particular qubit.

        If provided, store the result in the given output location.
        """
        self._add_statement(
            ast.QuantumMeasurementStatement(
                measure=ast.QuantumMeasurement(ast.Identifier(qubit.name)),
                target=optional_ast(self, output_location),
            )
        )
        return self

    def pragma(self, command: str) -> Program:
        """Add a pragma instruction."""
        if len(self.stack) != 1:
            raise RuntimeError("Pragmas must be global")
        self._add_statement(ast.Pragma(command))
        return self

    def _do_assignment(self, var: AstConvertible, op: str, value: AstConvertible) -> None:
        """Helper function for variable assignment operations."""
        if isinstance(var, classical_types.DurationVar):
            value = convert_float_to_duration(value)
        var_ast = to_ast(self, var)
        if isinstance(var_ast, ast.IndexExpression):
            assert isinstance(var_ast.collection, ast.Identifier)
            var_ast = ast.IndexedIdentifier(name=var_ast.collection, indices=[var_ast.index])
        self._add_statement(
            ast.ClassicalAssignment(
                var_ast,
                ast.AssignmentOperator[op],
                to_ast(self, value),
            )
        )

    def do_expression(self, expression: AstConvertible) -> Program:
        """Add a statement which evaluates a given expression without assigning the output."""
        self._add_statement(ast.ExpressionStatement(to_ast(self, expression)))
        return self

    def set(
        self,
        var: classical_types._ClassicalVar | classical_types.OQIndexExpression,
        value: AstConvertible,
    ) -> Program:
        """Set a variable value."""
        self._do_assignment(var, "=", value)
        return self

    def increment(self, var: classical_types._ClassicalVar, value: AstConvertible) -> Program:
        """Increment a variable value."""
        self._do_assignment(var, "+=", value)
        return self

    def decrement(self, var: classical_types._ClassicalVar, value: AstConvertible) -> Program:
        """Decrement a variable value."""
        self._do_assignment(var, "-=", value)
        return self

    def mod_equals(self, var: classical_types.IntVar, value: AstConvertible) -> Program:
        """In-place update of a variable to be itself modulo value."""
        assert isinstance(var, classical_types.IntVar)
        self._do_assignment(var, "%=", value)
        return self

    def annotate(self, keyword: str, command: Optional[str] = None) -> Program:
        """Add an annotation to the next statement."""
        self._state.annotations.append(ast.Annotation(keyword, command))
        return self


class MergeCalStatementsPass(QASMVisitor[None]):
    """Merge adjacent CalibrationStatement ast nodes."""

    def visit_Program(self, node: ast.Program, context: None = None) -> None:
        node.statements = self.process_statement_list(node.statements)
        self.generic_visit(node, context)

    def visit_ForInLoop(self, node: ast.ForInLoop, context: None = None) -> None:
        node.block = self.process_statement_list(node.block)
        self.generic_visit(node, context)

    def visit_WhileLoop(self, node: ast.WhileLoop, context: None = None) -> None:
        node.block = self.process_statement_list(node.block)
        self.generic_visit(node, context)

    def visit_BranchingStatement(self, node: ast.BranchingStatement, context: None = None) -> None:
        node.if_block = self.process_statement_list(node.if_block)
        node.else_block = self.process_statement_list(node.else_block)
        self.generic_visit(node, context)

    def visit_CalibrationStatement(
        self, node: ast.CalibrationStatement, context: None = None
    ) -> None:
        node.body = self.process_statement_list(node.body)
        self.generic_visit(node, context)

    def visit_SubroutineDefinition(
        self, node: ast.SubroutineDefinition, context: None = None
    ) -> None:
        node.body = self.process_statement_list(node.body)
        self.generic_visit(node, context)

    def process_statement_list(
        self, statements: list[ast.Statement | ast.Pragma]
    ) -> list[ast.Statement | ast.Pragma]:
        new_list = []
        cal_stmts = []
        for stmt in statements:
            if isinstance(stmt, ast.CalibrationStatement):
                cal_stmts.extend(stmt.body)
            else:
                if cal_stmts:
                    new_list.append(ast.CalibrationStatement(body=cal_stmts))
                    cal_stmts = []
                new_list.append(stmt)

        if cal_stmts:
            new_list.append(ast.CalibrationStatement(body=cal_stmts))

        return new_list
